from oibot.mixin import MixinProtocol


class RecallMessageMixin:
    async def recall_private_message(
        self: MixinProtocol, *, openid: str, message_id: str
    ) -> None:
        return await self(
            "DELETE",
            f"/v2/users/{openid}/messages/{message_id}",
            headers={"Authorization": f"QQBot {await self.get_app_access_token()}"},
        )

    async def recall_group_message(
        self: MixinProtocol, *, group_openid: str, message_id: str
    ) -> None:
        return await self(
            "DELETE",
            f"/v2/groups/{group_openid}/messages/{message_id}",
            headers={"Authorization": f"QQBot {await self.get_app_access_token()}"},
        )

    async def recall_channel_message(
        self: MixinProtocol, *, channel_id: str, message_id: str, hide_tip: bool = False
    ) -> None:
        return await self(
            "DELETE",
            f"/channels/{channel_id}/messages/{message_id}",
            headers={"Authorization": f"QQBot {await self.get_app_access_token()}"},
            params={"htdetip": hide_tip},
        )

    async def recall_guild_message(
        self: MixinProtocol, *, guild_id: str, message_id, hide_tip: bool = False
    ) -> None:
        return await self(
            "DELETE",
            f"/dms/{guild_id}/messages/{message_id}",
            headers={"Authorization": f"QQBot {await self.get_app_access_token()}"},
            params={"htdetip": hide_tip},
        )



from datetime import datetime
from typing import ClassVar, Literal

from oibot.bot import OiBot
from oibot.event import Context, Event
from oibot.mixin.send_message import (
    Ark,
    Embed,
    Keyboard,
    Markdown,
    Media,
    MessageReference,
    SendMessageResponse,
)


class GroupAddRobotEvent(Event):
    __slots__ = ("timestamp", "group_openid", "op_member_openid")

    event_type: ClassVar[Literal["GROUP_ADD_ROBOT"]] = "GROUP_ADD_ROBOT"

    timestamp: datetime
    group_openid: str
    op_member_openid: str

    def __init__(self, bot: OiBot, ctx: Context) -> None:
        d = ctx["d"]

        self.timestamp = d["timestamp"]
        self.group_openid = d["group_openid"]
        self.op_member_openid = d["op_member_openid"]

        super().__init__(bot, ctx)

    async def reply(
        self,
        content: str | None = None,
        markdown: Markdown | None = None,
        keyboard: Keyboard | None = None,
        embed: Embed | None = None,
        ark: Ark | None = None,
        media: Media | None = None,
        message_reference: MessageReference | None = None,
        event_id: str | None = None,
        msg_id: str | None = None,
        msg_seq: int = 0,
    ) -> SendMessageResponse:
        return await self.bot.send_group_message(
            group_openid=self.group_openid,
            content=content,
            markdown=markdown,
            keyboard=keyboard,
            embed=embed,
            ark=ark,
            media=media,
            message_reference=message_reference,
            event_id=event_id,
            msg_id=msg_id,
            msg_seq=msg_seq,
        )



from datetime import datetime
from typing import ClassVar, Literal

from oibot.bot import OiBot
from oibot.event import Context, Event


class GroupDelRobotEvent(Event):
    __slots__ = ("timestamp", "group_openid", "op_member_openid")

    event_type: ClassVar[Literal["GROUP_DEl_ROBOT"]] = "GROUP_DEl_ROBOT"

    timestamp: datetime
    group_openid: str
    op_member_openid: str

    def __init__(self, bot: OiBot, ctx: Context) -> None:
        d = ctx["d"]

        self.timestamp = d["timestamp"]
        self.group_openid = d["group_openid"]
        self.op_member_openid = d["op_member_openid"]

        super().__init__(bot, ctx)



import asyncio
from functools import partial, reduce, wraps
from operator import and_, or_
from typing import Any, Awaitable, Callable, Coroutine, Literal


def ensure_async(
    func: Callable[..., Any] | None = None, *, to_thread: bool = False
) -> Callable[..., Awaitable[Any]]:
    if func is None:
        return partial(ensure_async, to_thread=to_thread)

    if asyncio.iscoroutinefunction(func):
        return func

    if to_thread:

        @wraps(func)
        async def wrapper(*args, **kwargs) -> Any:
            return await asyncio.to_thread(func, *args, **kwargs)
    else:

        @wraps(func)
        async def wrapper(*args, **kwargs) -> Any:
            return func(*args, **kwargs)

    return wrapper


class Matcher:
    __slots__ = ("rule", "awaitable", "matchers", "operator")

    def __init__(
        self,
        rule: Callable[..., Any] = lambda *args, **kwargs: True,
        *,
        awaitable: bool = False,
        matchers: list["Matcher"] | None = None,
        operator: Literal["AND", "OR"] | None = None,
    ) -> None:
        self.rule = rule

        self.awaitable = (
            awaitable
            or asyncio.iscoroutinefunction(rule)
            or asyncio.iscoroutinefunction(self.__call__)
        )

        self.matchers = matchers or []
        self.operator = operator

    def __call__(
        self, *args, **kwargs
    ) -> dict[str, Any] | Coroutine[Any, Any, dict[str, Any]]:
        return self.rule(*args, **kwargs)

    def __and__(self, other: "Matcher") -> "Matcher":
        matchers = [
            *(self.matchers if self.operator == "AND" else [self]),
            *(other.matchers if other.operator == "AND" else [other]),
        ]

        async_matchers = [matcher for matcher in matchers if matcher.awaitable]
        sync_matchers = [matcher for matcher in matchers if not matcher.awaitable]

        if awaitable := async_matchers:

            async def wrapper(*args, **kwargs) -> dict[str, Any]:
                ctx = {}

                for matcher in sync_matchers:
                    if not (matched := matcher(*args, **kwargs)):
                        return {}

                    ctx |= matched if isinstance(matched, dict) else {"_": matched}

                try:
                    async for completed_task in asyncio.as_completed(
                        tasks := [
                            asyncio.create_task(
                                matcher(*args, **kwargs), eager_start=True
                            )
                            for matcher in async_matchers
                        ]
                    ):
                        if not (matched := await completed_task):
                            return {}

                        ctx |= matched if isinstance(matched, dict) else {"_": matched}

                finally:
                    for task in tasks:
                        if not task.done():
                            task.cancel()

                return ctx

        else:

            def wrapper(*args, **kwargs) -> dict[str, Any]:
                ctx = {}

                for matcher in sync_matchers:
                    if not (matched := matcher(*args, **kwargs)):
                        return {}

                    ctx |= matched if isinstance(matched, dict) else {"_": matched}

                return ctx

        return Matcher(
            rule=wrapper, awaitable=awaitable, matchers=matchers, operator="AND"
        )

    def __or__(self, other: "Matcher") -> "Matcher":
        matchers = [
            *(self.matchers if self.operator == "OR" else [self]),
            *(other.matchers if other.operator == "OR" else [other]),
        ]

        async_matchers = [matcher for matcher in matchers if matcher.awaitable]
        sync_matchers = [matcher for matcher in matchers if not matcher.awaitable]

        if awaitable := async_matchers:

            async def wrapper(*args, **kwargs) -> dict[str, Any]:
                for matcher in sync_matchers:
                    if matched := matcher(*args, **kwargs):
                        return matched if isinstance(matched, dict) else {"_": matched}

                try:
                    async for completed_task in asyncio.as_completed(
                        tasks := [
                            asyncio.create_task(
                                matcher(*args, **kwargs), eager_start=True
                            )
                            for matcher in async_matchers
                        ]
                    ):
                        if matched := await completed_task:
                            return (
                                matched if isinstance(matched, dict) else {"_": matched}
                            )

                finally:
                    for task in tasks:
                        if not task.done():
                            task.cancel()

                return {}
        else:

            def wrapper(*args, **kwargs) -> dict[str, Any]:
                for matcher in sync_matchers:
                    if matched := matcher(*args, **kwargs):
                        return matched if isinstance(matched, dict) else {"_": matched}

                return {}

        return Matcher(
            rule=wrapper, awaitable=awaitable, matchers=matchers, operator="OR"
        )

    def __invert__(self) -> "Matcher":
        if not self.awaitable:

            def wrapper(*args, **kwargs) -> bool:
                return not self(*args, **kwargs)

        else:

            async def wrapper(*args, **kwargs) -> bool:
                return not await self(*args, **kwargs)

        return Matcher(rule=wrapper, awaitable=self.awaitable)

    async def match(self, *args, **kwargs) -> dict[str, Any]:
        if matched := (
            await self(*args, **kwargs) if self.awaitable else self(*args, **kwargs)
        ):
            return matched if isinstance(matched, dict) else {"_": matched}

        return {}

    @classmethod
    def all(cls, *matchers: "Matcher") -> "Matcher":
        return reduce(and_, matchers)

    @classmethod
    def any(cls, *matchers: "Matcher") -> "Matcher":
        return reduce(or_, matchers)



from datetime import datetime
from typing import ClassVar, Literal, NamedTuple

from oibot.bot import OiBot
from oibot.event import Context, Event


class FriendDelEvent(Event):
    __slots__ = ("timestamp", "openid", "author")

    class Author(NamedTuple):
        union_openid: str

    event_type: ClassVar[Literal["FRIEND_DEL"]] = "FRIEND_DEL"

    timestamp: datetime
    openid: str
    author: Author

    def __init__(self, bot: OiBot, ctx: Context) -> None:
        d = ctx["d"]

        self.timestamp = d["timestamp"]
        self.openid = d["openid"]
        self.author = self.Author(union_openid=d["author"]["union_openid"])

        super().__init__(bot, ctx)



import logging
from enum import StrEnum
from typing import ClassVar, TypedDict

from oibot.bot import OiBot


class EventType(StrEnum):
    AT_MESSAGE_CREATE = "AT_MESSAGE_CREATE"
    AUDIO_FINISH = "AUDIO_FINISH"
    AUDIO_OR_LIVE_CHANNEL_MEMBER_ENTER = "AUDIO_OR_LIVE_CHANNEL_MEMBER_ENTER"
    AUDIO_OR_LIVE_CHANNEL_MEMBER_EXIT = "AUDIO_OR_LIVE_CHANNEL_MEMBER_EXIT"
    AUDIO_START = "AUDIO_START"
    C2C_MESSAGE_CREATE = "C2C_MESSAGE_CREATE"
    C2C_MSG_RECEIVE = "C2C_MSG_RECEIVE"
    C2C_MSG_REJECT = "C2C_MSG_REJECT"
    CHANNEL_CREATE = "CHANNEL_CREATE"
    CHANNEL_DELETE = "CHANNEL_DELETE"
    CHANNEL_UPDATE = "CHANNEL_UPDATE"
    DIRECT_MESSAGE_CREATE = "DIRECT_MESSAGE_CREATE"
    DIRECT_MESSAGE_DELETE = "DIRECT_MESSAGE_DELETE"
    FORUM_POST_CREATE = "FORUM_POST_CREATE"
    FORUM_POST_DELETE = "FORUM_POST_DELETE"
    FORUM_PUBLISH_AUDIT_RESULT = "FORUM_PUBLISH_AUDIT_RESULT"
    FORUM_REPLY_CREATE = "FORUM_REPLY_CREATE"
    FORUM_REPLY_DELETE = "FORUM_REPLY_DELETE"
    FORUM_THREAD_CREATE = "FORUM_THREAD_CREATE"
    FORUM_THREAD_DELETE = "FORUM_THREAD_DELETE"
    FORUM_THREAD_UPDATE = "FORUM_THREAD_UPDATE"
    FRIEND_ADD = "FRIEND_ADD"
    FRIEND_DEL = "FRIEND_DEL"
    GROUP_ADD_ROBOT = "GROUP_ADD_ROBOT"
    GROUP_AT_MESSAGE_CREATE = "GROUP_AT_MESSAGE_CREATE"
    GROUP_DEL_ROBOT = "GROUP_DEL_ROBOT"
    GROUP_MSG_RECEIVE = "GROUP_MSG_RECEIVE"
    GROUP_MSG_REJECT = "GROUP_MSG_REJECT"
    GUILD_CREATE = "GUILD_CREATE"
    GUILD_DELETE = "GUILD_DELETE"
    GUILD_MEMBER_ADD = "GUILD_MEMBER_ADD"
    GUILD_MEMBER_REMOVE = "GUILD_MEMBER_REMOVE"
    GUILD_MEMBER_UPDATE = "GUILD_MEMBER_UPDATE"
    GUILD_UPDATE = "GUILD_UPDATE"
    INTERACTION_CREATE = "INTERACTION_CREATE"
    MESSAGE_AUDIT_PASS = "MESSAGE_AUDIT_PASS"
    MESSAGE_AUDIT_REJECT = "MESSAGE_AUDIT_REJECT"
    MESSAGE_CREATE = "MESSAGE_CREATE"
    MESSAGE_DELETE = "MESSAGE_DELETE"
    MESSAGE_REACTION_ADD = "MESSAGE_REACTION_ADD"
    MESSAGE_REACTION_REMOVE = "MESSAGE_REACTION_REMOVE"
    OFF_MIC = "OFF_MIC"
    ON_MIC = "ON_MIC"
    OPEN_FORUM_POST_CREATE = "OPEN_FORUM_POST_CREATE"
    OPEN_FORUM_POST_DELETE = "OPEN_FORUM_POST_DELETE"
    OPEN_FORUM_REPLY_CREATE = "OPEN_FORUM_REPLY_CREATE"
    OPEN_FORUM_REPLY_DELETE = "OPEN_FORUM_REPLY_DELETE"
    OPEN_FORUM_THREAD_CREATE = "OPEN_FORUM_THREAD_CREATE"
    OPEN_FORUM_THREAD_DELETE = "OPEN_FORUM_THREAD_DELETE"
    OPEN_FORUM_THREAD_UPDATE = "OPEN_FORUM_THREAD_UPDATE"
    PUBLIC_MESSAGE_DELETE = "PUBLIC_MESSAGE_DELETE"


class Context(TypedDict):
    d: dict
    id: str
    op: int
    s: int
    t: str


class Event:
    __slots__ = ("bot", "_d", "_id", "_op", "_s", "_t")

    _d: dict
    _id: str
    _op: int
    _s: int
    _t: str

    event_type: ClassVar[EventType]

    event: ClassVar[dict[str, type["Event"]]] = {}

    def __new__(cls, bot: OiBot, ctx: Context) -> "Event":
        return (
            event.__new__(event, bot, ctx)
            if (event := cls.dispatch(ctx)) is not cls
            else super().__new__(cls)
        )

    def __init__(self, bot: OiBot, ctx: Context) -> None:
        self.bot = bot

        self._d = ctx["d"]
        self._id = ctx["id"]
        self._op = ctx["op"]
        # self._s = ctx["s"]
        self._t = ctx["t"]

        logging.info(self.__repr__())

    def __init_subclass__(cls, *args, **kwargs) -> None:
        logging.debug(f"registered {cls} as type {cls._t}")

        Event.event[cls.event_type] = cls

    def __repr__(self) -> str:
        return f"""{self.__class__.__name__}({
            ", ".join(
                f"{item}={value}"
                for item in self.__slots__
                if (not item.startswith("__")) and (value := getattr(self, item, None))
            )
        })"""

    @classmethod
    def dispatch(cls, ctx: Context) -> type["Event"]:
        return cls.event.get(ctx["t"], cls)



from datetime import datetime
from typing import ClassVar, Literal, NamedTuple

from oibot.bot import OiBot
from oibot.event import Context, Event
from oibot.mixin.send_message import (
    Ark,
    Embed,
    Keyboard,
    Markdown,
    Media,
    MessageReference,
    SendMessageResponse,
)


class FriendAddEvent(Event):
    __slots__ = ("timestamp", "openid", "author")

    class Author(NamedTuple):
        union_openid: str

    event_type: ClassVar[Literal["FRIEND_ADD"]] = "FRIEND_ADD"

    timestamp: datetime
    openid: str
    author: Author

    def __init__(self, bot: OiBot, ctx: Context) -> None:
        d = ctx["d"]

        self.timestamp = d["timestamp"]
        self.openid = d["openid"]
        self.author = self.Author(union_openid=d["author"]["union_openid"])

        super().__init__(bot, ctx)

    async def reply(
        self,
        content: str | None = None,
        markdown: Markdown | None = None,
        keyboard: Keyboard | None = None,
        embed: Embed | None = None,
        ark: Ark | None = None,
        media: Media | None = None,
        message_reference: MessageReference | None = None,
        event_id: str | None = None,
        msg_id: str | None = None,
        msg_seq: int = 0,
    ) -> SendMessageResponse:
        return await self.bot.send_private_message(
            openid=self.author.union_openid,
            content=content,
            markdown=markdown,
            keyboard=keyboard,
            embed=embed,
            ark=ark,
            media=media,
            message_reference=message_reference,
            event_id=event_id,
            msg_id=msg_id,
            msg_seq=msg_seq,
        )



import asyncio
from datetime import datetime
from itertools import count
from typing import ClassVar, Literal, NamedTuple

from oibot.bot import OiBot
from oibot.event import Context, Event
from oibot.mixin.send_message import (
    Ark,
    Embed,
    Keyboard,
    Markdown,
    Media,
    MessageReference,
    SendMessageResponse,
)


class C2CMessageCreateEvent(Event):
    __slots__ = (
        "id",
        "content",
        "timestamp",
        "author",
        "attachments",
        "message_scene",
        "message_type",
        "msg_seq",
    )

    class Author(NamedTuple):
        id: str
        user_openid: str
        union_openid: str

    class Attachment(NamedTuple):
        content_type: str
        filename: str
        height: int
        size: int
        url: str
        width: int

    class MessageScene(NamedTuple):
        source: str
        callback_data: str

    event_type: ClassVar[Literal["C2C_MESSAGE_CREATE"]] = "C2C_MESSAGE_CREATE"

    id: str
    content: str
    timestamp: datetime
    author: Author
    attachments: list[Attachment]

    message_scene: MessageScene
    message_type: int

    def __init__(self, bot: OiBot, ctx: Context) -> None:
        d = ctx["d"]

        self.id = d["id"]
        self.content = d["content"]
        self.timestamp = d["timestamp"]

        self.author = self.Author(
            id=d["author"]["id"],
            user_openid=d["author"]["user_openid"],
            union_openid=d["author"]["union_openid"],
        )

        self.attachments = [
            self.Attachment(
                content_type=attachment["content_type"],
                filename=attachment["filename"],
                height=attachment["height"],
                size=attachment["size"],
                url=attachment["url"],
                width=attachment["width"],
            )
            for attachment in d.get("attachments", [])
        ]

        self.message_scene = self.MessageScene(
            source=d["message_scene"]["source"],
            callback_data=d["message_scene"]["callback_data"],
        )
        self.message_type = d["message_type"]

        super().__init__(bot, ctx)

        self.msg_seq = count()

    async def reply(
        self,
        content: str | None = None,
        markdown: Markdown | None = None,
        keyboard: Keyboard | None = None,
        embed: Embed | None = None,
        ark: Ark | None = None,
        media: Media | None = None,
        message_reference: MessageReference | None = None,
        event_id: str | None = None,
        msg_seq: int = 0,
    ) -> SendMessageResponse:
        return await self.bot.send_private_message(
            openid=self.author.union_openid,
            content=content,
            markdown=markdown,
            keyboard=keyboard,
            embed=embed,
            ark=ark,
            media=media,
            message_reference=message_reference,
            event_id=event_id,
            msg_id=self.id,
            msg_seq=msg_seq or next(self.msg_seq),
        )

    async def defer(
        self,
        content: str | None = None,
        markdown: Markdown | None = None,
        keyboard: Keyboard | None = None,
        embed: Embed | None = None,
        ark: Ark | None = None,
        media: Media | None = None,
        message_reference: MessageReference | None = None,
        event_id: str | None = None,
        msg_seq: int = 0,
    ) -> "C2CMessageCreateEvent":
        self.bot.plugin_manager.sessions[
            key := (
                self.bot.bot.get(),
                getattr(self, "group_openid", None),
                (
                    getattr(self, "author.member_openid", None)
                    or getattr(self, "author.user_openid", None)
                ),
            )
        ] = future = asyncio.get_running_loop().create_future()

        await self.reply(
            content=content,
            markdown=markdown,
            keyboard=keyboard,
            embed=embed,
            ark=ark,
            media=media,
            message_reference=message_reference,
            event_id=event_id,
            msg_seq=msg_seq,
        )

        try:
            return await future

        finally:
            self.bot.plugin_manager.sessions.pop(key, None)



import asyncio
from functools import _make_key
from typing import Any, Callable, TypedDict

from oibot.mixin import MixinProtocol


class AccessToken(TypedDict):
    access_token: str
    expires_in: str


def keep_alive(func: Callable[..., Any]) -> Callable[..., Any]:
    futures = {}

    async def wrapper(*args, **kwargs) -> Any:
        key = _make_key(args, kwargs, typed=False)

        if future := futures.get(key):
            return await future

        loop = asyncio.get_running_loop()

        futures[key] = future = loop.create_future()

        try:
            result: AccessToken = await func(*args, **kwargs)
            access_token = result["access_token"]
            future.set_result(access_token)

            loop.call_later(int(result["expires_in"]), lambda: futures.pop(key, None))

            return access_token

        except BaseException as e:
            futures.pop(key, None)

            future.set_exception(e)

            raise

    return wrapper


class AccessTokenMixin:
    @keep_alive
    async def get_app_access_token(
        self: MixinProtocol, app_id: str | None = None, app_secret: str | None = None
    ) -> str:
        return await self(
            "POST",
            "https://bots.qq.com/app/getAppAccessToken",
            json={
                "appId": app_id or self.bot.get().app_id,
                "clientSecret": app_secret or self.bot.get().app_secret,
            },
        )



import asyncio
import logging
import os
import sys
from contextlib import (
    AbstractAsyncContextManager,
    AbstractContextManager,
    AsyncExitStack,
    asynccontextmanager,
    contextmanager,
)
from functools import wraps
from importlib import import_module, reload
from inspect import (
    Parameter,
    isasyncgenfunction,
    isclass,
    iscoroutinefunction,
    isgeneratorfunction,
    signature,
)
from textwrap import dedent
from types import ModuleType, UnionType
from typing import (
    Annotated,
    Any,
    Awaitable,
    Callable,
    NamedTuple,
    Union,
    get_args,
    get_origin,
)

from oibot.bot import Bot, OiBot
from oibot.event import Context, Event
from oibot.matcher import Matcher


class Plugin:
    class Executor(NamedTuple):
        func: Callable[..., Any]

        async def __call__(self, event: Event) -> Any:
            try:
                return await self.func(event)

            except RuntimeWarning as e:
                if reply := getattr(event, "reply", None):
                    if len(args := e.args) > 1:
                        for arg in args:
                            await reply(arg)

                    elif msg := str(e):
                        await reply(msg)

                    elif doc := self.func.__doc__ or self.func.__module__.__doc__:
                        await reply(dedent(doc))

            except RuntimeError as e:
                logging.exception(e)

    __slots__ = ("module", "executors")

    def __init__(self, module: ModuleType) -> None:
        self.module = module

        self.executors = [
            handler
            for handler in vars(module).values()
            if isinstance(handler, self.Executor)
            and handler.func.__module__ == module.__name__
        ]

    async def __call__(self, event: Event) -> None:
        try:
            async with asyncio.TaskGroup() as tg:
                for executor in self.executors:
                    tg.create_task(executor(event))

        except* RuntimeWarning as e:
            if reply := getattr(event, "reply", None):
                if args := e.args:
                    for arg in args:
                        await reply(arg)

                else:
                    await reply(str(e))

        except* RuntimeError as e:
            logging.exception(e)


class Dependency:
    __slots__ = ("dependency", "signature", "use_cache")

    def __init__(
        self, dependency: Callable[..., Any], *, use_cache: bool = True
    ) -> None:
        self.dependency = dependency
        self.signature = signature(dependency)

        self.use_cache = use_cache

    @classmethod
    def provide(cls, dependency: Callable[..., Any], *, use_cache: bool = True) -> Any:
        return cls(dependency, use_cache=use_cache)


class PluginManager:
    __slots__ = ("bot", "plugins", "sessions")

    def __init__(self, bot: OiBot) -> None:
        self.bot = bot

        self.plugins: dict[str, Plugin] = {}
        self.sessions: dict[tuple[Bot, str | None, str | None], asyncio.Future] = {}

    async def __call__(self, ctx: Context) -> Any:
        event = Event(self.bot, ctx)

        try:
            if future := self.sessions.get(
                (
                    self.bot.bot.get(),
                    getattr(event, "group_openid", None),
                    (
                        getattr(event, "author.member_openid", None)
                        or getattr(event, "author.user_openid", None)
                    ),
                )
            ):
                future.set_result(event)

            async with asyncio.TaskGroup() as tg:
                for plugin in self.plugins.values():
                    tg.create_task(plugin(event))

        except* Exception as eg:
            logging.exception(eg)

    def import_from(self, plugins: str) -> None:
        def load_module(module_path: str) -> None:
            module_name = (
                os.path.relpath(module_path, os.path.dirname(plugins))
                .removesuffix(".py")
                .replace(os.sep, ".")
            )

            try:
                if module_name in sys.modules:
                    logging.info(f"reloading plugin [{module_name}]")

                    module = reload(sys.modules[module_name])

                else:
                    module = import_module(module_name)

                self.plugins[module_name] = plugin = Plugin(module)

                logging.info(f"loaded plugin [{module_name}]")

                if not plugin.executors:
                    logging.warning(
                        f"unloaded plugin [{module_name}] due to missing handlers"
                    )

                    self.plugins.pop(module_name, None)

            except Exception as e:
                logging.exception(f"failed to load plugin [{module_name}]: {e}")

                self.plugins.pop(module_name, None)

        if os.path.isdir(plugins):
            for root, _, files in os.walk(plugins):
                for file in files:
                    if file.endswith(".py") and not file.startswith("_"):
                        load_module(os.path.join(root, file))

        elif (
            os.path.isfile(plugins)
            and plugins.endswith(".py")
            and not os.path.basename(plugins).startswith("_")
        ):
            load_module(plugins)


def on(
    matchers: Matcher | Callable[..., bool | Awaitable[bool]] | None = None,
) -> Callable[..., Any]:
    def annotation_event_type(annotation: Any) -> tuple[type[Event], ...]:
        if get_origin(annotation) in (Annotated, Union, UnionType):
            return tuple(
                arg
                for arg in get_args(annotation)
                if isclass(arg) and issubclass(arg, Event)
            )

        elif isclass(annotation) and issubclass(annotation, Event):
            return (annotation,)

        else:
            return ()

    async def resolve_dependency(
        event: Event,
        dependency: Dependency,
        dependency_cache: dict[Callable[..., Any], asyncio.Future],
        stack: AsyncExitStack,
    ) -> Any:
        if dependency.use_cache:
            if future := dependency_cache.get(dependency.dependency):
                return await future

            dependency_cache[dependency.dependency] = future = (
                asyncio.get_running_loop().create_future()
            )

        try:
            func = dependency.dependency
            sign = dependency.signature

            kwargs: dict[str, Any] = {}
            tasks: dict[str, asyncio.Task] = {}

            async with asyncio.TaskGroup() as tg:
                for param_name, param in sign.parameters.items():
                    if isinstance(param.default, Dependency):
                        tasks[param_name] = tg.create_task(
                            resolve_dependency(
                                event, param.default, dependency_cache, stack
                            )
                        )

                    elif param.annotation is OiBot:
                        kwargs[param_name] = event.bot

                    elif isinstance(event, annotation_event_type(param.annotation)):
                        kwargs[param_name] = event

                    elif param.default is not Parameter.empty:
                        kwargs[param_name] = param.default

                    elif param.kind in (
                        Parameter.VAR_POSITIONAL,
                        Parameter.VAR_KEYWORD,
                    ):
                        pass

                    else:
                        raise ValueError(
                            f"cannot resolve dependency for parameter '{param_name}' "
                            f"in function '{func.__name__}'. "
                            f"parameter must have either a default value, be an Event, or be a Dependency"
                        )

            kwargs.update({k: v.result() for k, v in tasks.items()})

            if isclass(func) and issubclass(func, AbstractAsyncContextManager):
                result = await stack.enter_async_context(func(**kwargs))

            elif isclass(func) and issubclass(func, AbstractContextManager):
                result = stack.enter_context(func(**kwargs))

            elif isasyncgenfunction(func):
                result = await stack.enter_async_context(
                    asynccontextmanager(func)(**kwargs)
                )

            elif isgeneratorfunction(func):
                result = stack.enter_context(contextmanager(func)(**kwargs))

            elif iscoroutinefunction(func):
                result = await func(**kwargs)

            else:
                result = func(**kwargs)

            if dependency.use_cache:
                future.set_result(result)

            return result

        except BaseException as e:
            future.set_exception(e)

            raise e

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        sign = signature(func)

        event_type = tuple(
            event
            for param in sign.parameters.values()
            for event in annotation_event_type(param.annotation)
        )

        matcher = (
            matchers
            if isinstance(matchers, Matcher)
            else Matcher(matchers or (lambda _: True))
        )

        if any(
            param
            for param in sign.parameters.values()
            if isinstance(param.default, Dependency) or param.annotation is OiBot
        ):

            @wraps(func)
            async def wrapper(event: Event, **kwargs) -> Any:
                if isinstance(event, event_type) and (
                    matched := await matcher.match(event)
                ):
                    func_kwargs: dict[str, Any] = {}
                    tasks: dict[str, asyncio.Task] = {}
                    dependency_cache: dict[Callable[..., Any], asyncio.Future] = {}

                    async with AsyncExitStack() as stack:
                        async with asyncio.TaskGroup() as tg:
                            for param_name, param in sign.parameters.items():
                                if isinstance(param.default, Dependency):
                                    tasks[param_name] = tg.create_task(
                                        resolve_dependency(
                                            event,
                                            param.default,
                                            dependency_cache,
                                            stack,
                                        )
                                    )

                                elif param.annotation is OiBot:
                                    func_kwargs[param_name] = event.bot

                                elif isinstance(
                                    event, annotation_event_type(param.annotation)
                                ):
                                    func_kwargs[param_name] = event

                                elif param.default is not Parameter.empty:
                                    func_kwargs[param_name] = param.default

                                elif param_name in matched:
                                    func_kwargs[param_name] = matched.pop(param_name)

                                elif param.kind is Parameter.VAR_POSITIONAL:
                                    pass

                                elif param.kind is Parameter.VAR_KEYWORD:
                                    func_kwargs.update(matched)

                                else:
                                    raise ValueError(
                                        f"cannot resolve dependency for parameter '{param_name}' "
                                        f"in function '{func.__name__}'. "
                                        f"parameter must have either a default value, be an Event, or be a Dependency."
                                    )

                        return await func(
                            **{
                                **kwargs,
                                **func_kwargs,
                                **{k: v.result() for k, v in tasks.items()},
                            }
                        )

        else:
            if not (
                (param := next(iter(sign.parameters.values()), None))
                and annotation_event_type(param.annotation)
            ):
                raise TypeError(
                    f"function '{func.__name__}' must accept an Event object as its first parameter."
                )

            if any(p.kind == Parameter.VAR_KEYWORD for p in sign.parameters.values()):

                @wraps(func)
                async def wrapper(event: Event, **kwargs) -> Any:
                    if isinstance(event, event_type) and (
                        matched := await matcher.match(event)
                    ):
                        return await func(event, **{**kwargs, **matched})

            else:

                @wraps(func)
                async def wrapper(event: Event, **kwargs) -> Any:
                    if isinstance(event, event_type) and (
                        matched := await matcher.match(event)
                    ):
                        return await func(
                            event,
                            **{
                                **kwargs,
                                **{
                                    k: v
                                    for k, v in matched.items()
                                    if k in sign.parameters
                                },
                            },
                        )

        return Plugin.Executor(wrapper)

    return decorator



from enum import IntEnum
from typing import Literal, NotRequired, TypedDict

from oibot.mixin import MixinProtocol
from oibot.mixin.upload_file import Media


class MsgType(IntEnum):
    PLAINTEXT = 0
    MIX = 1
    MARKDOWN = 2
    ARK = 3
    EMBED = 4
    MEDIA = 7


class MarkdownParam(TypedDict):
    key: str
    value: list[str]


class Markdown(TypedDict):
    content: NotRequired[str]
    custom_template_id: NotRequired[str]
    params: NotRequired[MarkdownParam]


class ButtonRenderData(TypedDict):
    label: str
    visited_label: str
    style: Literal[0, 1]


class ButtonActionPermission(TypedDict):
    type: Literal[0, 1, 2, 3]
    specify_user_ids: NotRequired[list[str]]
    specify_role_ids: NotRequired[list[str]]


class ButtonAction(TypedDict):
    type: Literal[0, 1, 2]
    data: str
    reply: NotRequired[bool]
    enter: NotRequired[bool]
    anchor: NotRequired[Literal[1]]
    click_limit: NotRequired[int]
    at_bot_show_channel_list: NotRequired[bool]
    unsupport_tips: str


class Button(TypedDict):
    id: NotRequired[str]
    render_data: ButtonRenderData
    action: ButtonAction


class Buttons(TypedDict):
    buttons: list[Button]


class Rows(TypedDict):
    rows: list


class Content(TypedDict):
    content: Rows


class Keyboard(TypedDict):
    id: NotRequired[str]
    content: NotRequired[str]


class Thumbnail(TypedDict):
    url: str


class EmbedField(TypedDict):
    name: str


class Embed(TypedDict):
    title: NotRequired[str]
    prompt: NotRequired[str]
    thumbnail: NotRequired[Thumbnail]
    fields: NotRequired[list[EmbedField]]


class ArkObjKV(TypedDict):
    key: str
    value: str


class ArkObj(TypedDict):
    obj_kv: list[ArkObjKV]


class ArkKV(TypedDict):
    key: str
    value: NotRequired[str]
    obj: NotRequired[list[ArkObj]]


class Ark(TypedDict):
    template_id: int
    kv: list


class MessageReference(TypedDict): ...


class SendMessageResponse(TypedDict):
    id: str
    timestamp: int


class SendMessageMixin:
    async def send_private_message(
        self: MixinProtocol,
        *,
        openid: str,
        content: str | None = None,
        markdown: Markdown | None = None,
        keyboard: Keyboard | None = None,
        embed: Embed | None = None,
        ark: Ark | None = None,
        media: Media | None = None,
        message_reference: MessageReference | None = None,
        event_id: str | None = None,
        msg_id: str | None = None,
        msg_seq: int = 0,
    ) -> SendMessageResponse:
        if markdown:
            msg_type = MsgType.MARKDOWN
        elif ark:
            msg_type = MsgType.ARK
        elif media:
            msg_type = MsgType.MEDIA
        else:
            msg_type = MsgType.PLAINTEXT

        file = await self.upload_private_file(openid=openid, **media) if media else None

        return await self(
            "POST",
            f"/v2/users/{openid}/messages",
            headers={"Authorization": f"QQBot {await self.get_app_access_token()}"},
            json={
                "content": content,
                "msg_type": msg_type,
                "markdown": markdown,
                "keyboard": keyboard,
                "embed": embed,
                "ark": ark,
                "media": file,
                "message_reference": message_reference,
                "event_id": event_id,
                "msg_id": msg_id,
                "msg_seq": msg_seq,
            },
        )

    async def send_group_message(
        self: MixinProtocol,
        *,
        group_openid: str,
        content: str | None = None,
        markdown: Markdown | None = None,
        keyboard: Keyboard | None = None,
        embed: Embed | None = None,
        ark: Ark | None = None,
        media: Media | None = None,
        message_reference: MessageReference | None = None,
        event_id: str | None = None,
        msg_id: str | None = None,
        msg_seq: int = 0,
    ) -> SendMessageResponse:
        if markdown:
            msg_type = MsgType.MARKDOWN
        elif ark:
            msg_type = MsgType.ARK
        elif media:
            msg_type = MsgType.MEDIA
        else:
            msg_type = MsgType.PLAINTEXT

        file = (
            await self.upload_group_file(group_openid=group_openid, **media)
            if media
            else None
        )

        return await self(
            "POST",
            f"/v2/groups/{group_openid}/messages",
            headers={"Authorization": f"QQBot {await self.get_app_access_token()}"},
            json={
                "content": content,
                "msg_type": msg_type,
                "markdown": markdown,
                "keyboard": keyboard,
                "embed": embed,
                "ark": ark,
                "media": file,
                "message_reference": message_reference,
                "event_id": event_id,
                "msg_id": msg_id,
                "msg_seq": msg_seq,
            },
        )



from base64 import b64encode
from enum import IntEnum
from typing import NotRequired, Self, TypedDict

from oibot.mixin import MixinProtocol


class FileType(IntEnum):
    IMAGE = 1
    VIDEO = 2
    VOICE = 3
    FILE = 4


class Media(dict):
    def __init__(
        self,
        file_type: FileType,
        url: str | None = None,
        srv_send_msg: bool = False,
        file_data: bytes | str | None = None,
    ) -> None:
        super().__init__(
            file_type=file_type,
            url=url,
            srv_send_msg=srv_send_msg,
            file_data=file_data
            and (
                b64encode(file_data).decode("utf-8")
                if isinstance(file_data, bytes)
                else file_data
            ),
        )

    @classmethod
    def image(
        cls,
        url: str | None = None,
        srv_send_msg: bool = False,
        file_data: bytes | str | None = None,
    ) -> Self:
        return cls(
            file_type=FileType.IMAGE,
            url=url,
            srv_send_msg=srv_send_msg,
            file_data=file_data,
        )

    @classmethod
    def video(
        cls,
        url: str | None = None,
        srv_send_msg: bool = False,
        file_data: bytes | str | None = None,
    ) -> Self:
        return cls(
            file_type=FileType.VIDEO,
            url=url,
            srv_send_msg=srv_send_msg,
            file_data=file_data,
        )

    @classmethod
    def voice(
        cls,
        url: str | None = None,
        srv_send_msg: bool = False,
        file_data: bytes | str | None = None,
    ) -> Self:
        return cls(
            file_type=FileType.VOICE,
            url=url,
            srv_send_msg=srv_send_msg,
            file_data=file_data,
        )

    @classmethod
    def file(
        cls,
        url: str | None = None,
        srv_send_msg: bool = False,
        file_data: bytes | str | None = None,
    ) -> Self:
        return cls(
            file_type=FileType.FILE,
            url=url,
            srv_send_msg=srv_send_msg,
            file_data=file_data,
        )


class UploadFileResponse(TypedDict):
    file_uuid: str
    file_info: str
    ttl: int
    id: NotRequired[str]


class UploadFileMixin:
    async def upload_private_file(
        self: MixinProtocol,
        *,
        openid: str,
        file_type: FileType,
        url: str | None = None,
        srv_send_msg: bool = False,
        file_data: bytes | str | None = None,
    ) -> UploadFileResponse:
        return await self(
            "POST",
            f"/v2/users/{openid}/files",
            headers={"Authorization": f"QQBot {await self.get_app_access_token()}"},
            json={
                "file_type": file_type,
                "url": url,
                "srv_send_msg": srv_send_msg,
                "file_data": file_data,
            },
        )

    async def upload_group_file(
        self: MixinProtocol,
        *,
        group_openid: str,
        file_type: FileType,
        url: str | None = None,
        srv_send_msg: bool = False,
        file_data: bytes | str | None = None,
    ) -> UploadFileResponse:
        return await self(
            "POST",
            f"/v2/groups/{group_openid}/files",
            headers={"Authorization": f"QQBot {await self.get_app_access_token()}"},
            json={
                "file_type": file_type,
                "url": url,
                "srv_send_msg": srv_send_msg,
                "file_data": file_data,
            },
        )



from contextvars import ContextVar
from typing import TYPE_CHECKING, Any, Literal, Optional, Protocol

if TYPE_CHECKING:
    from oibot.bot import Bot
    from oibot.mixin.send_message import (
        Ark,
        Embed,
        Keyboard,
        Markdown,
        Media,
        MessageReference,
        SendMessageResponse,
    )
    from oibot.mixin.upload_file import FileType, UploadFileResponse


class MixinProtocol(Protocol):
    bot: ContextVar["Bot"]

    async def __call__(
        self,
        method: Literal[
            "GET",
            "POST",
            "PUT",
            "DELETE",
            "PATCH",
            "HEAD",
            "OPTIONS",
            "TRACE",
            "CONNECT",
        ],
        url: str,
        **kwargs,
    ) -> Any:
        pass

    async def get_app_access_token(
        self, app_id: str | None = None, app_secret: str | None = None
    ) -> str: ...

    async def recall_private_message(self, *, openid: str, message_id: str) -> None: ...

    async def recall_group_message(
        self, *, group_openid: str, message_id: str
    ) -> None: ...

    async def recall_channel_message(
        self,
        *,
        channel_id: str,
        message_id: str,
        hide_tip: bool = False,
    ) -> None: ...

    async def recall_guild_message(
        self, *, guild_id: str, message_id, hide_tip: bool = False
    ) -> None: ...

    async def upload_private_file(
        self,
        *,
        openid: str,
        file_type: "FileType",
        url: str | None = None,
        srv_send_msg: bool = False,
        file_data: bytes | str | None = None,
    ) -> "UploadFileResponse": ...

    async def upload_group_file(
        self,
        *,
        group_openid: str,
        file_type: "FileType",
        url: str | None = None,
        srv_send_msg: bool = False,
        file_data: bytes | str | None = None,
    ) -> "UploadFileResponse": ...

    async def send_private_message(
        self,
        *,
        openid: str,
        content: str | None = None,
        markdown: Optional["Markdown"] = None,
        keyboard: Optional["Keyboard"] = None,
        embed: Optional["Embed"] = None,
        ark: Optional["Ark"] = None,
        media: Optional["Media"] = None,
        message_reference: Optional["MessageReference"] = None,
        event_id: str | None = None,
        msg_id: str | None = None,
        msg_seq: int = 0,
    ) -> "SendMessageResponse": ...



import asyncio
import logging
from contextvars import ContextVar
from enum import IntEnum
from http import HTTPStatus
from os import environ
from typing import Any, Iterable, Literal, NamedTuple, Optional

from aiohttp import ClientSession, web
from aiohttp.web_request import Request
from aiohttp.web_response import Response
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey

from oibot.mixin.access_token import AccessTokenMixin
from oibot.mixin.recall_message import RecallMessageMixin
from oibot.mixin.send_message import SendMessageMixin
from oibot.mixin.upload_file import UploadFileMixin


class OP(IntEnum):
    MESSAGE = 0
    VERIFICATION = 13


class Bot(NamedTuple):
    app_id: str
    app_token: str
    app_secret: str


class Singleton(type):
    instance: Optional["OiBot"] = None

    def __call__(cls, *args, **kwargs) -> "OiBot":
        if not (instance := cls.instance):
            cls.instance = instance = super().__call__(*args, **kwargs)

        return instance


class OiBot(
    AccessTokenMixin,
    RecallMessageMixin,
    SendMessageMixin,
    UploadFileMixin,
    metaclass=Singleton,
):
    __slots__ = ("app", "plugin_manager", "bot")

    def __init__(self, plugins: str | Iterable[str] | None = None, **kwargs) -> None:
        from oibot.plugin import PluginManager

        self.plugin_manager = plugin_manager = PluginManager(self)

        if isinstance(plugins, str):
            plugin_manager.import_from(plugins)

        elif isinstance(plugins, Iterable):
            for plugin in plugins:
                plugin_manager.import_from(plugin)

        self.app = app = web.Application(**kwargs)

        app.router.add_post(path="/", handler=self.handler)

        self.bot: ContextVar[Bot] = ContextVar("bot")

    async def __call__(
        self,
        method: Literal[
            "GET",
            "POST",
            "PUT",
            "DELETE",
            "PATCH",
            "HEAD",
            "OPTIONS",
            "TRACE",
            "CONNECT",
        ],
        url: str,
        **kwargs,
    ) -> Any:
        logging.debug(f"{method=} {url=} {kwargs=}")

        async with ClientSession(
            base_url="https://api.sgroup.qq.com",  # "https://sandbox.api.sgroup.qq.com"
        ) as session:
            async with session.request(method, url, **kwargs) as resp:
                resp.raise_for_status()

                return await resp.json()

    async def handler(
        self, request: Request, *, background_tasks: set[asyncio.Task] = set()
    ) -> Response:
        ctx = await request.json()

        logging.debug(ctx)

        self.bot.set(
            bot := Bot(
                app_id=request.query.get("id", environ["OIBOT_APP_ID"]),
                app_token=request.query.get("token", environ["OIBOT_APP_TOKEN"]),
                app_secret=request.query.get("secret", environ["OIBOT_APP_SECRET"]),
            )
        )

        match ctx["op"]:
            case OP.MESSAGE:
                background_tasks.add(
                    task := asyncio.create_task(self.plugin_manager(ctx))
                )

                task.add_done_callback(background_tasks.discard)

            case OP.VERIFICATION:
                logging.info("webhook verification request received")

                if not (secret := bot.app_secret):
                    raise ValueError("parameter `app_secret` must be specified")

                secret = secret.encode("utf-8")

                while len(secret) < 32:
                    secret *= 2

                d = ctx["d"]

                return web.json_response(
                    {
                        "plain_token": d["plain_token"],
                        "signature": (
                            Ed25519PrivateKey.from_private_bytes(secret[:32])
                            .sign(f"{d['event_ts']}{d['plain_token']}".encode("utf-8"))
                            .hex()
                        ),
                    }
                )

            case _:
                logging.warning(f"invalid type received {ctx=}")

        return web.Response(body=None, status=HTTPStatus.OK)

    def run(self, *args, **kwargs) -> None:
        web.run_app(self.app, *args, **kwargs)
